
#include "usb910conf.h"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>

#include "usbcdc.h"
#include "usb910.h"

//#define FCNT_DEBUG

#define poll    usbcdc_poll

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))


#ifdef FCNT_EXTENSION

#define FCNT_PORT       PORTB
#define FCNT_DDR        DDRB
#define FCNT_PIN        PB6

static uint8_t t0_ov_count;
static uint8_t t1_ov_count;

#define MAX_OV_COUNT	(F_CPU / 8192 / 10)	// 146 @ 12MHz 

static union fn {
	uint32_t dw;
	struct {
		uint8_t b0;
		uint8_t b1;
		uint8_t b2;
		uint8_t b3;
	};
} fn;

#ifdef FCNT_DEBUG
static uint8_t tcnt0l,tcnt0h,tcnt1; 
#endif

static inline void fcnt_start() {
	t0_ov_count = 0;
	t1_ov_count = 0;
	TCCR1B = 0; // stop clock
	TCCR1A = 0;
        TCNT1 = 0;
	TCCR0B = 0; // timer0 stop
	TCCR0A = (1<<TCW0); // 16 bit mode
        TCNT0H = 0;
        TCNT0L = 0;
	cli();
	TCCR0B = 6; // timer0 start @ T0 down edge
	TCCR1B = (5 + 1) | (1<<PSR1); // 1/32;
	sei();
}

static inline void fcnt_stop() {
	cli();
	TCCR1B = 0; // stop clock
	TCCR0B = 0; // timer0 stop
	sei();
}

static inline void fcnt_calibrate() {
	int8_t t0_tmp,t0_prev;
	int8_t t0_tmp_l,t0_prev_l;
	int8_t t1_tmp,t1_prev;
	int8_t stopping,i;
	union fn t0,t1;
	t0_prev = 0;
	t0_prev_l = 0;
	t1_prev = 0;
	stopping = 0;
	fcnt_start();
	for (;;) {
		t0_tmp_l = TCNT0L;
		t0_tmp = TCNT0H;
		t1_tmp = TCNT1;
		if (t0_prev < 0 && t0_tmp >= 0) {
			t0_ov_count++;
			if (stopping) break;
			if (t0_ov_count >= MAX_OV_COUNT)
				stopping = 1;
		}
		if (t1_prev < 0 && t1_tmp >= 0) {
			t1_ov_count++;
			if (stopping) break;
			if (t1_ov_count >= MAX_OV_COUNT)
				stopping = 2;
		}
		if (stopping == 1 || stopping == 2) {
			if (t0_tmp > 0 && t1_tmp > 0)
				stopping += 2;
		} if (stopping == 3) {
			if (t1_prev != t1_tmp) break;
		} if (stopping == 4) {
			if (t0_prev_l != t0_tmp_l) break;
		}
		t0_prev_l = t0_tmp_l;
		t0_prev = t0_tmp;
		t1_prev = t1_tmp;
		wdt_reset();
		poll();
	}
	fcnt_stop();
	t0.b0 = TCNT0L;
	t0.b1 = TCNT0H;
	t0.b2 = t0_ov_count;
	t0.b3 = 0;
	t1.b0 = TCNT1;
	t1.b1 = t1_ov_count;
	t1.b2 = 0;
	t1.b3 = 0;
#ifdef FCNT_DEBUG
	tcnt0l = TCNT0L;
	tcnt0h = TCNT0H;
	tcnt1 = TCNT1;
#endif

	if (t1_ov_count == 0) {
		fn.dw = 0;
		return;
	}
	/* 
	 * t1 / (F_CPU/32) = t0 / HZ 
	 *                :
	 * HZ = t0 * F_CPU / t1 / 32 
	 */
	for (i=0; i< 5; i++) {
		if (t0.dw < ( 0xffffffff / (F_CPU/4000)))
			break;
		t0.dw >>= 1;
	}
	t0.dw *= (F_CPU/4000);
	t0.dw /= t1.dw;
	for (;i < 5; i++) {
		if (t0.dw < ( 0xffffffff / (4000)))
			break;
		t0.dw >>= 1;
	}
	t0.dw *= 4000;
	t0.dw >>= (5-i);
	fn.dw = t0.dw;
}


int8_t fcnt_send(uint8_t size) {
	if (size >= 1 && ext_buf[0] == FCNT_START) {
		fcnt_calibrate();
		return 1;
	} else return -2;
}

int8_t fcnt_recv(uint8_t size) {
	uint8_t r;
	if (size >= 3) {
		r = 0;
		ext_buf[0] = fn.b0;
		ext_buf[1] = fn.b1;
		ext_buf[2] = fn.b2;
		r += 3;
#ifdef FCNT_DEBUG
		if (size >= 4) {
			ext_buf[3] = t1_ov_count;
			r++;
		}
		if (size >= 5) {
			ext_buf[4] = t0_ov_count;
			r++;
		}
		if (size >= 6) {
			ext_buf[5] = tcnt0l;
			r++;
		}
		if (size >= 7) {
			ext_buf[6] = tcnt0h;
			r++;
		}
		if (size >= 8) {
			ext_buf[7] = tcnt1;
			r++;
		}
#endif
		return r;
		return r;
	} else return -2;
}

#endif
